import 'package:flutter_riverpod/flutter_riverpod.dart';
import 'package:endlich/endlich.dart';
import 'package:riverpod_endlich/src/value_provider.dart';
import 'provider.dart';

import 'builder.dart';

abstract class AutoDisposeMachineStateProviderRef<State>
    implements AutoDisposeProviderRefBase {}

class AutoDisposeMachineStateProviderElement<State>
    extends AutoDisposeProviderElementBase<State>
    implements AutoDisposeMachineStateProviderRef<State> {
  AutoDisposeMachineStateProviderElement(
      AutoDisposeProviderBase<State> provider)
      : super(provider);
}

class AutoDisposeMachineStateProvider<Machine extends StateMachine<State>,
    State> extends AutoDisposeProviderBase<State> {
  AutoDisposeMachineStateProvider(this._create, {String? name}) : super(name);

  static const family = AutoDisposeMachineStateProviderFamilyBuilder();

  final Create<Machine, AutoDisposeChangeNotifierProviderRef<Machine>> _create;

  late final machine = AutoDisposeChangeNotifierProvider<Machine>(_create,
      name: (null != name ? '$name.machine' : null));

  @override
  ProviderBase<Object?> get providerToRefresh => machine.notifier;

  @override
  State create(AutoDisposeChangeNotifierProviderRef<Machine> ref) {
    final ele = ref as AutoDisposeMachineStateProviderElement<State>;
    final s = ele.watch(machine).state;
    return s;
  }

  @override
  AutoDisposeProviderElementBase<State> createElement() {
    return AutoDisposeMachineStateProviderElement<State>(this);
  }

  @override
  bool updateShouldNotify(State previousState, State newState) {
    return true;
  }

  Override overrideWithProvider(
    AutoDisposeMachineStateProvider<Machine, State> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: this, override: this);
      setup(origin: machine, override: machine);
      setup(origin: machine.notifier, override: provider.machine.notifier);
    });
  }

  @override
  void setupOverride(SetupOverride setup) {
    setup(origin: this, override: this);
    setup(origin: machine, override: machine);
    setup(origin: machine.notifier, override: machine.notifier);
  }

  Override overrideWithValue(Machine value) {
    return ProviderOverride((setup) {
      setup(origin: this, override: this);
      setup(
        origin: machine,
        override: machine,
      );
      setup(
        origin: machine.notifier,
        override: ValueProvider<Machine>(value),
      );
    });
  }
}

class AutoDisposeMachineStateProviderFamily<Machine extends StateMachine<State>,
        State, Arg>
    extends Family<State, Arg,
        AutoDisposeMachineStateProvider<Machine, State>> {
  AutoDisposeMachineStateProviderFamily(this._create, {String? name})
      : super(name);

  final FamilyCreate<Machine, AutoDisposeChangeNotifierProviderRef<Machine>,
      Arg> _create;

  @override
  AutoDisposeMachineStateProvider<Machine, State> create(
    Arg argument,
  ) {
    final provider = AutoDisposeMachineStateProvider<Machine, State>(
      (ref) => _create(ref, argument),
      name: name,
    );

    registerProvider(provider.machine, argument);
    registerProvider(provider.machine.notifier, argument);

    return provider;
  }

  /// Overrides the behavior of a family for a part of the application.
  ///
  /// {@macro riverpod.overideWith}
  Override overrideWithProvider(
    AutoDisposeMachineStateProvider<Machine, State> Function(Arg argument)
        override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider, override: provider);
        setup(origin: provider.machine, override: provider.machine);
        setup(
            origin: provider.machine.notifier,
            override: override(arg).machine.notifier);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.machine, override: provider.machine);
    setup(
        origin: provider.machine.notifier, override: provider.machine.notifier);
  }
}
