import 'package:flutter_riverpod/flutter_riverpod.dart';
import 'package:riverpod/riverpod.dart';
import 'package:endlich/endlich.dart';

import 'builder.dart';
import 'value_provider.dart';

abstract class MachineStateProviderRef<State> implements ProviderRefBase {}

class MachineStateProviderElement<State> extends ProviderElementBase<State>
    implements MachineStateProviderRef<State> {
  MachineStateProviderElement(ProviderBase<State> provider) : super(provider);
}

class MachineStateProvider<Machine extends StateMachine<State>, State>
    extends AlwaysAliveProviderBase<State> {
  MachineStateProvider(this._create, {String? name}) : super(name);

  static const family = MachineStateProviderFamilyBuilder();

  static const autoDispose = AutoDisposeMachineStateProviderBuilder();

  static const autoDisposeFamily =
      AutoDisposeMachineStateProviderFamilyBuilder();

  final Create<Machine, ChangeNotifierProviderRef<Machine>> _create;

  late final machine = ChangeNotifierProvider<Machine>(_create,
      name: (null != name ? '$name.machine' : null));

  @override
  State create(MachineStateProviderRef<State> ref) {
    final ele = ref as MachineStateProviderElement<State>;
    final s = ele.watch(machine).state;
    return s;
  }

  @override
  ProviderBase<Object?> get providerToRefresh => machine.notifier;

  @override
  void setupOverride(SetupOverride setup) {
    setup(origin: this, override: this);
    setup(origin: machine, override: machine);
    setup(origin: machine.notifier, override: machine.notifier);
  }

  Override overrideWithValue(Machine value) {
    return ProviderOverride((setup) {
      setup(origin: this, override: this);
      setup(
        origin: machine,
        override: machine,
      );
      setup(
        origin: machine.notifier,
        override: ValueProvider<Machine>(value),
      );
    });
  }

  @override
  ProviderElementBase<State> createElement() {
    return MachineStateProviderElement<State>(this);
  }

  @override
  bool updateShouldNotify(State previousState, State newState) {
    return true;
  }

  Override overrideWithProvider(
    MachineStateProvider<Machine, State> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: this, override: this);
      setup(origin: machine, override: machine);
      setup(origin: machine.notifier, override: provider.machine.notifier);
    });
  }
}

class MachineStateProviderFamily<Machine extends StateMachine<State>, State,
    Arg> extends Family<State, Arg, MachineStateProvider<Machine, State>> {
  MachineStateProviderFamily(this._create, {String? name}) : super(name);

  final FamilyCreate<Machine, ChangeNotifierProviderRef<Machine>, Arg> _create;

  @override
  MachineStateProvider<Machine, State> create(
    Arg argument,
  ) {
    final provider = MachineStateProvider<Machine, State>(
      (ref) => _create(ref, argument),
      name: name,
    );

    registerProvider(provider.machine, argument);
    registerProvider(provider.machine.notifier, argument);

    return provider;
  }

  Override overrideWithProvider(
    MachineStateProvider<Machine, State> Function(Arg argument) override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider, override: provider);
        setup(origin: provider.machine, override: provider.machine);
        setup(
            origin: provider.machine.notifier,
            override: override(arg).machine.notifier);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.machine, override: provider.machine);
    setup(
        origin: provider.machine.notifier, override: provider.machine.notifier);
  }
}
