import 'package:flutter_test/flutter_test.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod_endlich/riverpod_endlich.dart';

import 'stuff.dart';

void main() {
  group('ChangeNotifierProvider.autoDispose.family', () {
    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () async {
        final provider =
            MachineStateProvider.autoDisposeFamily<TestMachine, Test, int>(
                (ref, _) => TestMachine());
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider(0).machine.notifier).state, isFoo());
        expect(container.read(provider(0)), isFoo());
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine.notifier)
          ]),
        );
        expect(root.getAllProviderElementsInOrder(), isEmpty);
      });

      test('when using provider.overrideWithProvider', () async {
        final provider =
            MachineStateProvider.autoDisposeFamily<TestMachine, Test, int>(
                (ref, _) => TestMachine());
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(
            (value) => MachineStateProvider.autoDispose(
                (ref) => TestMachine(init: (m) => m.toBar(0))),
          ),
        ]);

        expect(container.read(provider(0).machine.notifier).state, isBar());
        expect(container.read(provider(0)), isBar());
        expect(root.getAllProviderElementsInOrder(), isEmpty);
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine.notifier)
          ]),
        );
      });
    });
  });
}
