import 'package:flutter_test/flutter_test.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod_endlich/src/provider.dart';

import 'stuff.dart';

void main() {
  group('ChangeNotifierProvider.autoDispose', () {
    test('can be refreshed', () async {
      var result = TestMachine();
      final container = createContainer();
      final provider =
          MachineStateProvider.autoDispose<TestMachine, Test>((ref) => result);

      expect(container.read(provider), isFoo());
      expect(container.read(provider.machine.notifier), result);

      result = TestMachine(init: (m) => m.toBaz(0));
      expect(container.refresh(provider), isBaz());

      expect(container.read(provider), isBaz());
      expect(container.read(provider.machine.notifier), result);
    });

    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () {
        final provider = MachineStateProvider.autoDispose<TestMachine, Test>(
            (ref) => TestMachine());
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider.machine.notifier).state, isFoo());
        expect(container.read(provider), isFoo());
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine.notifier)
          ]),
        );
        expect(root.getAllProviderElements(), isEmpty);
      });

      test('when using provider.overrideWithValue', () {
        final provider = MachineStateProvider.autoDispose<TestMachine, Test>(
            (ref) => TestMachine());
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithValue(TestMachine(init: (m) => m.toBar(0))),
        ]);

        expect(container.read(provider.machine.notifier).state, isBar());
        expect(container.read(provider), isBar());
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine.notifier)
          ]),
        );
        expect(root.getAllProviderElements(), isEmpty);
      });

      test('when using provider.overrideWithProvider', () {
        final provider = MachineStateProvider.autoDispose<TestMachine, Test>(
            (ref) => TestMachine());
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(
            MachineStateProvider.autoDispose<TestMachine, Test>(
                (ref) => TestMachine(init: (m) => m.toBar(0))),
          ),
        ]);

        expect(container.read(provider.machine.notifier).state, isBar());
        expect(container.read(provider), isBar());
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine.notifier)
          ]),
        );
        expect(root.getAllProviderElements(), isEmpty);
      });
    });
  });
}
