import 'package:endlich/endlich.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/mockito.dart';
import 'package:riverpod/riverpod.dart';

part 'stuff.freezed.dart';

class TestMachine extends StateMachine<Test> {
  TestMachine({
    this.init,
    this.onFoo,
    this.onBar,
    this.onBaz,
  });

  void Function(TestMachine machine)? init;
  void Function(Foo foo, StateConfig<Test> config, TestMachine machine)? onFoo;
  void Function(Bar bar, StateConfig<Test> config, TestMachine machine)? onBar;
  void Function(Baz baz, StateConfig<Test> config, TestMachine machine)? onBaz;

  void toFoo(int i) => transition((config) {
        final s = Foo(
          i: i,
          toBar: () => toBar(i + 1),
          next: () => toBar(i + 1),
        );
        onFoo?.call(s, config, this);
        return s;
      });

  void toBar(int i) => transition((config) {
        final s = Bar(
            i: i,
            toFoo: () => toFoo(i + 1),
            toBaz: () => toBaz(i + 1),
            next: () => toBaz(i + 1));
        onBar?.call(s, config, this);
        return s;
      });

  void toBaz(int i) => transition((config) {
        final s = Baz(
            i: i,
            toFoo: () => toFoo(i + 1),
            toBar: () => toBar(i + 1),
            next: () => toFoo(i + 1));
        onBaz?.call(s, config, this);
        return s;
      });

  @override
  void createInitialState() {
    if (null != init) {
      init!(this);
      return;
    }
    toFoo(0);
  }
}

@freezed
class Test with _$Test {
  factory Test.foo({
    required int i,
    required void Function() toBar,
    required void Function() next,
  }) = Foo;
  factory Test.bar({
    required int i,
    required void Function() toFoo,
    required void Function() toBaz,
    required void Function() next,
  }) = Bar;

  factory Test.baz({
    required int i,
    required void Function() toFoo,
    required void Function() toBar,
    required void Function() next,
  }) = Baz;
}

class StateListener extends Mock {
  void call(Test value);
}

class ControllerListenerMock extends Mock {
  void call(StateMachine<Test>? value);
}

typedef VerifyOnly = VerificationResult Function<T>(
  Mock mock,
  T matchingInvocations,
);

TypeMatcher<StateMachine<Test>> isAValueOf(Test value) {
  return isA<StateMachine<Test>>().having((s) => s.state, 'state.value', value);
}

TypeMatcher<Test> isFoo([int? i]) {
  final a = isA<Foo>();
  return null == i ? a : a.having((p0) => p0.i, 'check int', equals(i));
}

TypeMatcher<Test> isBar([int? i]) {
  final a = isA<Bar>();
  return null == i ? a : a.having((p0) => p0.i, 'check int', equals(i));
}

TypeMatcher<Test> isBaz([int? i]) {
  final a = isA<Baz>();
  return null == i ? a : a.having((p0) => p0.i, 'check int', equals(i));
}

TypeMatcher<StateMachine<Test>> isOfTest<T extends Test>() {
  return isA<StateMachine<Test>>().having((s) => s.state, 'state', isA<T>());
}

/// Syntax sugar for:
///
/// ```dart
/// verify(mock()).called(1);
/// verifyNoMoreInteractions(mock);
/// ```
VerifyOnly get verifyOnly {
  final verification = verify;

  return <T>(mock, invocation) {
    final result = verification(invocation);
    result.called(1);
    verifyNoMoreInteractions(mock);
    return result;
  };
}

ProviderContainer createContainer({
  ProviderContainer? parent,
  List<Override> overrides = const [],
  List<ProviderObserver>? observers,
}) {
  final container = ProviderContainer(
    parent: parent,
    overrides: overrides,
    observers: observers,
  );
  addTearDown(container.dispose);
  return container;
}
